from django import forms
from django.contrib.auth.forms import AuthenticationForm

from birder.models import Monitor, Project, User


class DateInput(forms.DateInput):
    input_type = "date"


class LoginForm(AuthenticationForm):
    username = forms.EmailField(label="Email", widget=forms.TextInput(attrs={"autofocus": True}))

    class Meta:
        model = User
        fields = ("username", "password")


class MonitorForm(forms.ModelForm):
    class Meta:
        model = Monitor
        fields = (
            "project",
            "environment",
            "name",
            "position",
            "description",
            "notes",
            "custom_icon",
            "strategy",
            "configuration",
        )


class ProjectForm(forms.ModelForm):
    class Meta:
        model = Project
        fields = ("name", "environments", "public", "bitcaster_url", "icon")



import datetime

MINUTE = 1
MINUTES_30 = 30
MINUTES_PER_HOUR = 60
MINUTES_PER_DAY = 24 * MINUTES_PER_HOUR


def get_start_of_day(dt: datetime) -> datetime:
    return dt.replace(hour=0, minute=0, second=0, microsecond=0)


def format_minutes_as_time(minute: int) -> str:
    if minute < 1 or minute > MINUTES_PER_DAY:
        raise ValueError(f"{minute} is not a valid minute")

    hour = (minute - 1) // MINUTES_PER_HOUR
    minute_of_hour = (minute - 1) % MINUTES_PER_HOUR
    return f"{hour:02}:{minute_of_hour:02}"



import os
from typing import Any

import click
from click import Context
from tabulate import tabulate

import birder


@click.group()
@click.version_option(version=birder.VERSION)
def cli(**kwargs: Any) -> None:
    import django

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "birder.config.settings")
    django.setup()


@cli.command(name="list")
@click.pass_context
def list_(ctx: Context, **kwargs: Any) -> None:
    """List all existing monitors."""
    from birder.models import Monitor

    data = Monitor.objects.values(
        "id",
        "project__name",
        "name",
        "strategy",
        "active",
    )
    table = tabulate(data, [], tablefmt="simple")
    click.echo(table)


@cli.command(name="check")
@click.argument("monitor_id", type=int, required=False)
@click.option("-a", "--all", "_all", type=int, is_flag=True)
@click.pass_context
def trigger(ctx: Context, monitor_id: int, _all: bool = False, **kwargs: Any) -> None:
    """Run selected check."""
    from birder.models import BaseCheck, Monitor

    if _all and monitor_id:
        raise click.UsageError("--")
    ok = click.style("\u2714", fg="green")
    ko = click.style("\u2716", fg="red")
    if _all:
        for monitor in Monitor.objects.select_related("project", "environment").order_by(
            "project__name", "environment", "name"
        ):
            if monitor.strategy.mode == BaseCheck.LOCAL_TRIGGER:
                res = monitor.run()
                status = ok if res else ko
                click.echo(
                    f"{monitor.project.name[:20]:<22} | "
                    f"{monitor.environment.name[:15]:<17} | "
                    f"{monitor.name[:20]:<22} | "
                    f"{status}"
                )
    else:
        monitor = Monitor.objects.get(id=monitor_id)
        if monitor.strategy.mode == BaseCheck.LOCAL_TRIGGER:
            monitor.run()


@cli.command()
@click.argument("monitor_id", type=int)
@click.pass_context
def refresh(ctx: Context, monitor_id: int, **kwargs: Any) -> None:
    """Force UI refresh."""
    from birder.models import Monitor
    from birder.ws.utils import notify_ui

    monitor = Monitor.objects.get(id=monitor_id)
    notify_ui("update", monitor=monitor)


@cli.command()
@click.pass_context
def reset(ctx: Context, **kwargs: Any) -> None:
    """Reset all checks."""
    from django.core.cache import cache

    cache.clear()


def main() -> None:
    cli(prog_name=birder.NAME, obj={}, max_content_width=100)



from typing import Any
from urllib.parse import parse_qsl, urlparse

from .amqp import AmqpCheck
from .base import BaseCheck
from .celery import CeleryCheck
from .ftp import FtpCheck
from .http import HttpCheck
from .json import JsonCheck
from .ldap import LDAPCheck
from .memcache import MemCacheCheck
from .mysql import MySQLCheck
from .passive import HealthCheck
from .pg import PostgresCheck
from .redis import RedisCheck
from .registry import registry
from .s3 import S3Check
from .smtp import SMTPCheck
from .ssh import SSHCheck
from .tcp import TCPCheck
from .xml import XMLCheck

registry.register(AmqpCheck)
registry.register(CeleryCheck)
registry.register(FtpCheck)
registry.register(HealthCheck)
registry.register(HttpCheck)
registry.register(JsonCheck)
registry.register(LDAPCheck)
registry.register(MemCacheCheck)
registry.register(MySQLCheck)
registry.register(PostgresCheck)
registry.register(RedisCheck)
registry.register(S3Check)
registry.register(SMTPCheck)
registry.register(SSHCheck)
registry.register(TCPCheck)
registry.register(XMLCheck)


def parse_uri(uri: str) -> dict[str, str | Any]:
    o = urlparse(uri)
    cfg = {
        **dict(parse_qsl(o.query)),
        "hostname": o.hostname,
        "host": o.hostname,
        "scheme": o.scheme,
        "username": o.username,
        "password": o.password,
        "address": f"{o.scheme}://{o.hostname}{o.path}",
    }
    if o.port:
        cfg["port"] = o.port
    return cfg


def parser(uri: str) -> tuple[type[BaseCheck], dict[str, str | Any]]:
    checker: type[BaseCheck] = registry.checker_from_url(uri)
    url_config = parse_uri(uri)
    config = checker.clean_config({**checker.config_class.DEFAULTS, **url_config})
    chk = checker(configuration=config)
    return checker, chk.config



from ..settings import env

# Social Auth settings.
SOCIAL_AUTH_SECRET = env.str("AZURE_CLIENT_SECRET")
SOCIAL_AUTH_TENANT_ID = env("AZURE_TENANT_ID")
SOCIAL_AUTH_KEY = env.str("AZURE_CLIENT_KEY")
SOCIAL_AUTH_GOOGLE_OAUTH2_KEY = env("GOOGLE_CLIENT_ID")
SOCIAL_AUTH_GOOGLE_OAUTH2_SECRET = env("GOOGLE_CLIENT_SECRET")


SOCIAL_AUTH_REDIRECT_IS_HTTPS = env("SOCIAL_AUTH_REDIRECT_IS_HTTPS")
SOCIAL_AUTH_ADMIN_USER_SEARCH_FIELDS = [
    "username",
    "first_name",
    "last_name",
    "email",
]
SOCIAL_AUTH_JSONFIELD_ENABLED = True
SOCIAL_AUTH_URL_NAMESPACE = "social"
SOCIAL_AUTH_STRATEGY = "social_django.strategy.DjangoStrategy"


SOCIAL_AUTH_PIPELINE = (
    "social_core.pipeline.social_auth.social_details",
    "social_core.pipeline.social_auth.social_uid",
    "social_core.pipeline.social_auth.social_user",
    "social_core.pipeline.user.get_username",
    "social_core.pipeline.social_auth.associate_by_email",
    "social_core.pipeline.user.create_user",
    "social_core.pipeline.social_auth.associate_user",
    "social_core.pipeline.social_auth.load_extra_data",
    "social_core.pipeline.user.user_details",
    "birder.utils.pipeline.configure_user",
)

SOCIAL_AUTH_SANITIZE_REDIRECTS = False
SOCIAL_AUTH_USERNAME_IS_FULL_EMAIL = True
SOCIAL_AUTH_RAISE_EXCEPTIONS = env("SOCIAL_AUTH_RAISE_EXCEPTIONS")

SOCIAL_LOGIN_URL = env("SOCIAL_AUTH_LOGIN_URL")

USER_FIELDS = ["username", "email", "first_name", "last_name"]
SOCIAL_AUTH_GOOGLE_OAUTH2_WHITELISTED_DOMAINS = env("SOCIAL_AUTH_WHITELISTED_DOMAINS")
SOCIAL_AUTH_LOGIN_ERROR_URL = "/auth/error/"
SOCIAL_AUTH_BACKEND_ERROR_URL = "/auth/error/"



from typing import Any

import pymysql
from django import forms
from django.core.validators import MinValueValidator

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField


class MySQLConfig(ConfigForm):
    host = forms.CharField(required=True)
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=3306)
    database = forms.CharField(required=False)
    user = forms.CharField(required=False)
    password = WriteOnlyField(required=False)
    connect_timeout = forms.IntegerField(initial=2)


class MySQLCheck(BaseCheck):
    icon = "mysql.svg"
    pragma = ["mysql"]
    config_class = MySQLConfig
    address_format = "{host}:{port}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("database"):
            cfg["database"] = cfg.get("path", "")
        if not cfg.get("password"):
            cfg["password"] = cfg.get("password", "")
        if not cfg.get("user"):
            cfg["user"] = cfg.get("username", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            conn = pymysql.connect(**self.config)
            cursor = conn.cursor()
            cursor.execute("SELECT 1 FROM DUAL;")
            return True
        except pymysql.err.OperationalError as e:
            if raise_error:
                raise CheckError("MySQL check failed") from e
        return False



import json

from asgiref.sync import async_to_sync
from channels.generic.websocket import JsonWebsocketConsumer

GROUP = "ALL"


class CheckConsumer(JsonWebsocketConsumer):
    def connect(self) -> None:
        async_to_sync(self.channel_layer.group_add)(GROUP, self.channel_name)
        self.accept()
        self.send(text_data=json.dumps({"type": "connect"}))

    def disconnect(self, close_code: str) -> None:
        async_to_sync(self.channel_layer.group_discard)(GROUP, self.channel_name)



from typing import Any

from django.core.cache import cache
from django.http.request import HttpRequest


def birder(request: HttpRequest) -> dict[str, Any]:
    return {"birder": {"system": {"last_check": cache.get("system:last_check")}}}



import dramatiq.brokers.redis
import valkey

from ..settings import env

dramatiq.brokers.redis.redis = valkey

DRAMATIQ_VALKEY_URL = env("VALKEY_URL")
DRAMATIQ_BROKER = {
    "BROKER": "dramatiq.brokers.redis.RedisBroker",
    "OPTIONS": {
        "connection_pool": valkey.ConnectionPool.from_url(DRAMATIQ_VALKEY_URL),
    },
    "MIDDLEWARE": [
        # "dramatiq.middleware.Prometheus",
        "dramatiq.middleware.AgeLimit",
        "dramatiq.middleware.TimeLimit",
        "dramatiq.middleware.Callbacks",
        "dramatiq.middleware.Retries",
        "django_dramatiq.middleware.DbConnectionsMiddleware",
        "django_dramatiq.middleware.AdminMiddleware",
    ],
}

DRAMATIQ_TASKS_DATABASE = "default"
DRAMATIQ_AUTODISCOVER_MODULES = ["tasks"]



from smtplib import SMTP, SMTPException

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField


class SMTPConfig(ConfigForm):
    host = forms.CharField(required=True, help_text="Server hostname or IP Address")
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=587)
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)
    username = forms.CharField(required=False)
    password = WriteOnlyField(required=False)


class SMTPCheck(BaseCheck):
    icon = "smtp.svg"
    pragma = ["smtp"]
    config_class = SMTPConfig
    address_format = "{host}:{port}"

    def check(self, raise_error: bool = False) -> bool:
        try:
            config = {**self.config}
            username, password = config.pop("username"), config.pop("password")
            server = SMTP(**config)
            server.starttls()
            if username and password:
                server.login(username, password)
            server.quit()
            return True
        except (ConnectionRefusedError, SMTPException) as e:
            if raise_error:
                raise CheckError("SMTP check failed") from e
        return False



from typing import TYPE_CHECKING, Any

from django import template
from django.template import Context
from django.templatetags.static import static
from django.urls import reverse

if TYPE_CHECKING:
    from django.http import HttpRequest


register = template.Library()

MAX_ERRORS = 9


@register.simple_tag()
def status(m: str) -> str:
    if m:
        return static(f"images/{m}.svg")
    return static("images/question.svg")


@register.simple_tag()
def number(m: str) -> str:
    try:
        c = int(m)
        if c > MAX_ERRORS:
            return static("images/numbers/#.svg")
        return static(f"images/numbers/{c}.svg")
    except ValueError:
        return static("images/numbers/0.svg")


@register.simple_tag(takes_context=True)
def absolute_url(context: Context, *args: Any, **kwargs: Any) -> str:
    request: HttpRequest = context["request"]
    name = args[0]
    args = args[1:]
    return request.build_absolute_uri(reverse(name, args=args, kwargs=kwargs))



import json
from datetime import datetime, timedelta
from typing import Any

from django.contrib.auth.views import LoginView as LoginView_
from django.http.request import HttpRequest
from django.http.response import HttpResponse
from django.shortcuts import get_object_or_404, render
from django.utils.safestring import mark_safe
from django.views.generic import DetailView, TemplateView

from birder.checks import BaseCheck
from birder.config import settings
from birder.forms import LoginForm
from birder.models import Monitor, Project
from birder.utils.dates import format_minutes_as_time, get_start_of_day
from birder.ws.utils import notify_ui


class CommonContextMixin:
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        kwargs["active_view"] = self.__class__.__name__
        kwargs["sso_enabled"] = bool(settings.SOCIAL_AUTH_GOOGLE_OAUTH2_KEY)

        return super().get_context_data(**kwargs)


class IndexView(CommonContextMixin, TemplateView):
    template_name = "index.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        kwargs["projects"] = Project.objects.filter(public=True)
        return super().get_context_data(**kwargs)


class ProjectView(CommonContextMixin, TemplateView):
    template_name = "project.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        project = Project.objects.get(pk=self.kwargs.get("project"))
        if selection := self.kwargs.get("env"):
            env = project.environments.get(name=selection)
        else:
            env = project.environments.first()

        kwargs["selected_env"] = env
        filters = {"environment": env}
        kwargs["project"] = project
        monitors = Monitor.objects.filter(**filters).order_by("position", "name")
        kwargs["monitors"] = monitors
        kwargs["environments"] = project.environments.order_by("name")
        for monitor in monitors:
            notify_ui("update", monitor=monitor)
        return super().get_context_data(**kwargs)


class MonitorDetail(CommonContextMixin, DetailView):
    template_name = "monitor.html"
    queryset = Monitor.objects.all()

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        from birder.db import DataStore

        now = get_start_of_day(datetime.now())
        bits = DataStore(self.object).get_all_entries(now)
        group_by = 5
        data = [sum(bits[i : i + group_by]) for i in range(0, len(bits), group_by)]
        kwargs["title"] = "{} - {}".format(
            now.strftime("%H:%M"), (now + timedelta(hours=(len(bits) // 60) - 1)).strftime("%H:%M")
        )
        kwargs["data"] = data
        kwargs["group_by"] = group_by
        kwargs["labels"] = mark_safe(  # noqa: S308
            json.dumps([format_minutes_as_time(i) for i in list(range(1, len(bits) + 1, group_by))])
        )
        return super().get_context_data(**kwargs)


class LoginView(CommonContextMixin, LoginView_):
    template_name = "login.html"
    form_class = LoginForm


def trigger(request: HttpRequest, pk: str, token: str) -> HttpResponse:
    m: Monitor = get_object_or_404(Monitor, pk=pk)
    if m.token != token:
        return HttpResponse("Invalid Token", status=403)
    if m.strategy.mode != BaseCheck.REMOTE_INVOCATION:
        return HttpResponse("Check not enabled for remote call", status=400)
    m.get()
    return HttpResponse("Ok")


def error_400(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/400.html", {"error_code": 400, "message": "Bad Request"}, status=400)


def error_403(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/403.html", {"error_code": 403, "message": "Forbidden"}, status=403)


def error_404(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/404.html", {"error_code": 404, "message": "Page not found"}, status=404)


def error_500(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/500.html", {"error_code": 500, "message": "Server Error"}, status=500)
